# Contains the get_orbits function which categorizes clusters into orbits and
# orientation families

from .utilities import BiMap
from .atoms import Atom
from .orbit import Orbit, get_geometrical_radius, get_maximum_distance
from .orientation_family import OrientationFamily
from ..io.logging import logger

import numpy as np

logger = logger.getChild('get_orbits')


# This is the interface accessible for cluster_space
def get_orbits(cluster_list, atom_list, rotation_matrices, translation_vectors,
               permutations, prim):
    """Generate a list of the orbits for the clusters in a supercell
    configuration.

    This method requires as input a list of the clusters in a supercell
    configuration as well as a set of symmetry operations (rotations and
    translations). From this information it will generate a list of the orbits,
    i.e. the set of symmetry inequivalent clusters each associated with its
    respective set of equivalent clusters.

    Parameters
    ----------
    cluster_list : BiMap object
        a list of clusters
    atom_list : BiMap object
        a list of atoms in a supercell
    rotation_matrices : list of NumPy (3,3) arrays
        rotational symmetries to be imposed (e.g., from spglib)
    translation_vectors : list of NumPy (3) arrays
        translational symmetries to be imposed (e.g., from spglib)
    permutations : list of permutations
        lookup table for permutations
    prim : hiPhive Atoms object
        primitive structure

    Returns
    -------
    list of Orbits objs
        orbits associated with the list of input clusters

    """

    logger.debug('Preparing input...')
    atoms = prepare_atoms(atom_list)
    clusters = prepare_clusters(cluster_list)
    rotations = prepare_rotations(rotation_matrices)
    translations = prepare_translations(translation_vectors)
    permutations = prepare_permutations(permutations)
    cell = prim.cell
    basis = prim.basis

    logger.debug('Creating permutation map...')
    permutation_map, extended_atoms = \
        get_permutation_map(atoms, rotations, translations, basis)

    logger.debug('Creating orbits...')
    orbits = _get_orbits(permutation_map, extended_atoms, clusters, basis,
                         cell, rotations, permutations)

    return orbits


# All prepares are in case we changes some interface stuff

def prepare_permutations(permutations):
    perms = BiMap()
    for p in permutations:
        perms.append(tuple(p))
    return perms


def prepare_atoms(atom_list):
    atoms = BiMap()
    for atom in atom_list:
        atoms.append(atom)
    return atoms


def prepare_clusters(cluster_list):
    clusters = BiMap()
    for cluster in cluster_list:
        clusters.append(tuple(cluster))
    return clusters


def prepare_rotations(rotation_matrices):
    return rotation_matrices


def prepare_translations(translation_vectors):
    return translation_vectors


def get_permutation_map(atoms, rotations, translations, basis):

    extended_atoms = atoms.copy()

    permutation_map = np.zeros((len(atoms), len(rotations)), dtype=int)

    scaled_positions = [atom.spos(basis) for atom in extended_atoms]

    for sym_index, (R, T) in enumerate(zip(rotations, translations)):
        for atom_index, spos in enumerate(scaled_positions):

            new_spos = np.dot(R, spos) + T
            new_atom = Atom.spos_to_atom(new_spos, basis)

            if new_atom not in extended_atoms:
                extended_atoms.append(new_atom)

            mapped_atom_index = extended_atoms.index(new_atom)
            permutation_map[atom_index, sym_index] = mapped_atom_index

    return permutation_map, extended_atoms


# The internal function doing the outer loop over orbits
def _get_orbits(permutation_map, extended_atoms, clusters,
                basis, cell,
                rotations, permutations):
    cluster_is_found = [False] * len(clusters)
    orbits = []
    for cluster_index, cluster in enumerate(clusters):
        if cluster_is_found[cluster_index]:
            continue

        orbit = Orbit()

        cluster_atoms = [extended_atoms[i] for i in cluster]

        positions = [atom.pos(basis, cell) for atom in cluster_atoms]
        orbit.radius = get_geometrical_radius(positions)
        orbit.maximum_distance = get_maximum_distance(positions)
        orbit.order = len(cluster)

        populate_orbit(orbit, permutations, clusters,
                       cluster,
                       permutation_map, extended_atoms,
                       cluster_is_found)
        orbits.append(orbit)

#            bar.tick()
    return orbits


# Takes a cluster and generates all equivalent, translated clusters
def generate_translated_clusters(cluster, extended_atoms):
    transformed_cluster_atoms = [extended_atoms[i] for i in cluster]
    tested_offsets = set()
    for atom in transformed_cluster_atoms:
        offset = atom.offset
        if offset in tested_offsets:
            continue
        else:
            tested_offsets.add(offset)
        translated_cluster = []
        for atom in transformed_cluster_atoms:
            new_offset = np.subtract(atom.offset, offset)
            new_atom = Atom(atom.site, new_offset)
            translated_cluster.append(extended_atoms.index(new_atom))
        yield tuple(translated_cluster)


# Here is the actual categorization
def populate_orbit(orbit, permutations, clusters,
                   cluster,
                   permutation_map, extended_atoms,
                   cluster_is_found):
    for sym_index in range(permutation_map.shape[1]):

        of = OrientationFamily(sym_index)

        transformed_cluster = permutation_map[cluster, sym_index]

        for translated_cluster in generate_translated_clusters(
                transformed_cluster, extended_atoms):

            argsort = tuple(np.argsort(translated_cluster))
            perm_index = permutations.index(argsort)

            translated_cluster = tuple(sorted(translated_cluster))
            translated_cluster_index = clusters.index(translated_cluster)

            if cluster == translated_cluster:
                if (sym_index, perm_index) not in orbit.eigensymmetries:
                    orbit.eigensymmetries.append((sym_index, perm_index))

            if not cluster_is_found[translated_cluster_index]:
                cluster_is_found[translated_cluster_index] = True
                of.cluster_indices.append(translated_cluster_index)
                of.permutation_indices.append(perm_index)

        if len(of.cluster_indices) > 0:
            orbit.orientation_families.append(of)

    return orbit



"""
The ``io.shengBTE`` module provides functions for reading and writing data
files in shengBTE format.

Todo
----
These functions have not really been tested, i.e. write a fc3 with
write_shengBTE_fc3 and pass the file to shengBTE

"""

import numpy as np
from ase.geometry import find_mic
from itertools import product


def read_shengBTE_fc3(filename):
    """ Read shengBTE fc3 file.

    Parameters
    ----------
    filename : str
        input file name

    Returns
    -------
    list
        list with shengBTE block entries, where an entry consist of
        [i, j, k, cell_pos2, cell_pos3, fc3]
    """
    lines_per_fc_block = 32

    fc3_data_shengBTE = []
    with open(filename, 'r') as f:
        lines = f.readlines()
        num_fcs = int(lines[0])
        lind = 2
        for i in range(1, num_fcs+1):
            # sanity check
            assert int(lines[lind]) == i, (int(lines[lind]), i)

            # read cell poses
            cell_pos2 = np.array([float(fld) for fld in lines[lind+1].split()])
            cell_pos3 = np.array([float(fld) for fld in lines[lind+2].split()])

            # read basis indices
            i, j, k = (int(fld) for fld in lines[lind+3].split())

            # read fc
            fc3_ijk = np.zeros((3, 3, 3))
            for n in range(27):
                x, y, z = [int(fld) - 1 for fld in lines[lind+4+n].split()[:3]]
                fc3_ijk[x, y, z] = float(lines[lind+4+n].split()[-1])

            # append entry
            entry = [i, j, k, cell_pos2, cell_pos3, fc3_ijk]
            fc3_data_shengBTE.append(entry)
            lind += lines_per_fc_block

    return fc3_data_shengBTE


def write_shengBTE_fc3(filename, fcs, prim):
    """Write third order force constants file in shengBTE format.

    Parameters
    -----------
    filename: str
        input file name
    fcs : ForceConstants
        forceconstants must contain atoms object, and it must be a repeated
        from the atoms_prim
    prim: ASE Atoms object
        primitive configuration (equivalent to ``POSCAR`` file used in the
        shengBTE calculation)
    """

    fc3_data_shengBTE = _fcs_to_shengBTE_data(fcs, prim)

    with open(filename, 'w') as f:
        f.write('{}\n\n'.format(len(fc3_data_shengBTE)))

        for index, fc3_row in enumerate(fc3_data_shengBTE, start=1):
            i, j, k, cell_pos2, cell_pos3, fc3_ijk = fc3_row

            f.write('{:5d}\n'.format(index))

            f.write((3*'{:14.10f}'+'\n').format(*cell_pos2))
            f.write((3*'{:14.10f}'+'\n').format(*cell_pos3))
            f.write((3*'{:5d}'+'\n').format(i, j, k))

            for x, y, z in product(range(3), repeat=3):
                f.write((3*' {:}').format(x+1, y+1, z+1))
                f.write('    {:14.10f}\n'.format(fc3_ijk[x, y, z]))
            f.write('\n')


def _fcs_to_shengBTE_data(fcs, prim, cutoff=np.inf):
    """ Produces a list containing fcs in shengBTE format """

    atoms = fcs.atoms
    indices, offsets = [], []
    for atom in atoms:
        spos = np.linalg.solve(prim.cell.T, atom.position)
        for index, spos_prim in enumerate(prim.get_scaled_positions()):
            offset = spos - spos_prim
            rounded_offset = np.round(offset)
            if not np.allclose(rounded_offset, offset, rtol=0, atol=1e-3):
                continue
            rounded_offset = rounded_offset.astype(np.int64)
            assert np.allclose(rounded_offset, offset)
            indices.append(index)
            offsets.append(rounded_offset)
            break
        else:
            raise ValueError('prim not compatible with supercell')
    indices, offsets = np.array(indices), np.array(offsets)

    mic = np.zeros((len(atoms), len(atoms)), dtype=bool)
    for a0 in atoms:
        for a1 in atoms:
            dv = a1.position - a0.position
            tmp = find_mic([dv], atoms.cell)
            dv_mic, dr_mic = tmp[0][0], tmp[1][0]
            mic[a0.index, a1.index] = np.allclose(
                dv, dv_mic) and (dr_mic < cutoff)

    data = {}
    n_atoms = len(atoms)
    for i in range(n_atoms):
        for j in range(n_atoms):
            if not mic[i, j]:
                continue
            for k in range(n_atoms):
                if not (mic[i, k] and mic[j, k]):
                    continue

                i0 = indices[i]
                i1 = indices[j]
                i2 = indices[k]
                o0 = offsets[i]
                o1 = offsets[j] - o0
                o2 = offsets[k] - o0

                key = (i0, i1, i2) + tuple(o1) + tuple(o2)

                if key in data:
                    continue

                data[key] = fcs[i, j, k]

    new_data = []
    for k, fc in data.items():
        if np.linalg.norm(fc) < 1e-8:
            continue
        entry = [k[0] + 1, k[1] + 1, k[2] + 1]
        cell_pos1 = np.dot(k[3:6], prim.cell)
        cell_pos2 = np.dot(k[6:9], prim.cell)
        entry.extend([cell_pos1, cell_pos2])
        entry.append(fc)
        new_data.append(entry)

    return new_data



"""
scikit-learn is an excellent library for training linear models and provides a
large number of useful tools.

This module provides simplified interfaces for vaiours linear model regression
methods. These methods are set up in a way that work out of the box for typical
problems in cluster expansion and force constant potential construction. This
includes slight adjustments scitkit-learn default values.

If you would like more flexibility or extended functionality or ability to
fine-tune parameters that are not included in this interface, it is of course
possible to use scikit-learn directly.
More information about the sklearn linear models can be found at
http://scikit-learn.org/stable/modules/linear_model.html
"""

import numpy as np
from ..io.logging import logger
from collections import OrderedDict
from .split_bregman import fit_split_bregman

from sklearn.linear_model import (Lasso,
                                  LinearRegression,
                                  LassoCV,
                                  ElasticNet,
                                  ElasticNetCV,
                                  BayesianRidge,
                                  ARDRegression)
from sklearn.model_selection import ShuffleSplit
from sklearn.feature_selection import RFE, RFECV
from sklearn.preprocessing import StandardScaler


logger = logger.getChild('fit_methods')


def fit(X, y, fit_method, standardize=True, check_condition=True, **kwargs):
    """ Wrapper function for all available fit methods.

    Parameters
    -----------
    X : numpy.ndarray or list(list(float))
        fit matrix
    y : numpy.ndarray
        target array
    fit_method : str
        method to be used for training; possible choice are
        "least-squares", "lasso", "elasticnet", "bayesian-ridge", "ardr"
    standardize : bool
        whether or not to standardize the fit matrix before fitting
    check_condition : bool
        whether or not to carry out a check of the condition number

        N.B.: This can be sligthly more time consuming for larger
        matrices.

    Returns
    ----------
    results : dict
        parameters and other pertinent information
    """

    if fit_method not in available_fit_methods:
        msg = ['Fit method not available']
        msg += ['Please choose one of the following:']
        for key in available_fit_methods:
            msg += [' * ' + key]
        raise ValueError('\n'.join(msg))

    if check_condition:
        cond = np.linalg.cond(X)
        if cond > 1e10:
            logger.warning('Condition number is large, {}'.format(cond))

    if standardize:
        ss = StandardScaler(copy=False, with_mean=False, with_std=True)
        ss.fit_transform(X)  # change in place
        results = fit_methods[fit_method](X, y, **kwargs)
        ss.inverse_transform(X)  # change in place
        ss.transform(results['parameters'].reshape(1, -1)).reshape(-1,)
    else:
        results = fit_methods[fit_method](X, y, **kwargs)
    return results


def _fit_least_squares(X, y):
    """
    Returns the least-squares solution `a` to the linear problem `Xa=y`.

    This function is a wrapper to the `linalg.lstsq` function in NumPy.

    Parameters
    -----------
    X : numpy.ndarray
        fit matrix
    y : numpy.ndarray
        target array

    Returns
    ----------
    results : dict
        parameters
    """
    results = dict()
    results['parameters'] = np.linalg.lstsq(X, y, rcond=-1)[0]
    return results


def _fit_lasso(X, y, alpha=None, fit_intercept=False, **kwargs):
    """
    Return the solution `a` to the linear problem `Xa=y` obtained by using
    the LASSO method as implemented in scitkit-learn.

    LASSO optimizes the following problem::

        (1 / (2 * n_samples)) * ||y - Xw||^2_2 + alpha * ||w||_1

    If `alpha` is `None` this function will call the fit_lassoCV which attempts
    to find the optimal alpha via sklearn LassoCV class.

    Parameters
    -----------
    X : numpy.ndarray
        fit matrix
    y : numpy.ndarray
        target array
    alpha : float
        alpha value
    fit_intercept : bool
        center data or not, forwarded to sklearn

    Returns
    ----------
    results : dict
        parameters
    """
    if alpha is None:
        return _fit_lassoCV(X, y, fit_intercept=fit_intercept, **kwargs)
    else:
        lasso = Lasso(alpha=alpha, fit_intercept=fit_intercept, **kwargs)
        lasso.fit(X, y)
        results = dict()
        results['parameters'] = lasso.coef_
        return results


def _fit_lassoCV(X, y, alphas=None, fit_intercept=False, cv=10, n_jobs=-1,
                 **kwargs):
    """
    Returns the solution `a` to the linear problem `Xa=y` obtained by using
    the LassoCV method as implemented in scitkit-learn.

    Parameters
    -----------
    X : numpy.ndarray
        fit matrix
    y : numpy.ndarray
        target array
    alphas : list / array
        list of alpha values to be evaluated during regularization path
    fit_intercept : bool
        center data or not, forwarded to sklearn
    cv : int
        how many folds to carry out in cross-validation

    Returns
    -------
    results : dict
        parameters as well as
        `alpha_optimal` (alpha value that yields the lowest validation RMSE),
        `alpha_path` (all tested alpha values),
        `mse_path` (MSE for validation set for each alpha)
    """

    if alphas is None:
        alphas = np.logspace(-8, -0.3, 100)

    lassoCV = LassoCV(alphas=alphas, fit_intercept=fit_intercept, cv=cv,
                      n_jobs=n_jobs, **kwargs)
    lassoCV.fit(X, y)
    results = dict()
    results['parameters'] = lassoCV.coef_
    results['alpha_optimal'] = lassoCV.alpha_
    results['alpha_path'] = lassoCV.alphas_
    results['mse_path'] = lassoCV.mse_path_.mean(axis=1)
    return results


def _fit_elasticnet(X, y, alpha=None, fit_intercept=False, **kwargs):
    """
    Return the solution `a` to the linear problem `Xa=y` obtained by using
    the ElasticNet method as implemented in scitkit-learn.

    If `alpha` is `None` this function will call the fit_lassoCV which attempts
    to find the optimal alpha via sklearn ElasticNetCV class.

    Parameters
    -----------
    X : numpy.ndarray
        fit matrix
    y : numpy.ndarray
        target array
    alpha : float
        alpha value
    fit_intercept : bool
        center data or not, forwarded to sklearn

    Returns
    ----------
    results : dict
        parameters
    """
    if alpha is None:
        return _fit_elasticnetCV(X, y, fit_intercept=fit_intercept, **kwargs)
    else:
        elasticnet = ElasticNet(alpha=alpha, fit_intercept=fit_intercept,
                                **kwargs)
        elasticnet.fit(X, y)
        results = dict()
        results['parameters'] = elasticnet.coef_
        return results


def _fit_elasticnetCV(X, y, alphas=None, l1_ratio=None, fit_intercept=False,
                      cv=10, n_jobs=-1, **kwargs):
    """
    Returns the solution `a` to the linear problem `Xa=y` obtained by using
    the ElasticNetCV method as implemented in scitkit-learn.

    Parameters
    -----------
    X : numpy.ndarray
        fit matrix
    y : numpy.ndarray
        target array
    alphas : list or numpy.ndarray
        list of alpha values to be evaluated during regularization path
    l1_ratio : float or list(float)
        l1_ratio values to be evaluated during regularization path
    fit_intercept : bool
        center data or not, forwarded to sklearn
    cv : int
        how many folds to carry out in cross-validation

    Returns
    -------
    results : dict
        parameters as well as
        `alpha_optimal` (alpha value that yields the lowest validation RMSE),
        `alpha_path` (all tested alpha values),
        `l1_ratio_optmal` (alpha value that yields the lowest validation RMSE),
        `l1_ratio_path` (all tested `l1_ratio` values)
        `mse_path` (MSE for validation set for each alpha and `l1_ratio`)
    """

    if alphas is None:
        alphas = np.logspace(-8, -0.3, 100)
    if l1_ratio is None:
        l1_ratio = [1.0, 0.995, 0.99, 0.98, 0.97, 0.95, 0.925, 0.9, 0.85,
                    0.8, 0.75, 0.65, 0.5, 0.4, 0.25, 0.1]

    elasticnetCV = ElasticNetCV(alphas=alphas, l1_ratio=l1_ratio, cv=cv,
                                fit_intercept=fit_intercept, n_jobs=n_jobs,
                                **kwargs)
    elasticnetCV.fit(X, y)
    results = dict()
    results['parameters'] = elasticnetCV.coef_
    results['alpha_optimal'] = elasticnetCV.alpha_
    results['alpha_path'] = elasticnetCV.alphas_
    results['l1_ratio_path'] = elasticnetCV.l1_ratio
    results['l1_ratio_optimal'] = elasticnetCV.l1_ratio_
    results['mse_path'] = elasticnetCV.mse_path_.mean(axis=2)
    return results


def _fit_bayesian_ridge(X, y, fit_intercept=False, **kwargs):
    """
    Returns the solution `a` to the linear problem `Xa=y` obtained by using
    Bayesian ridge regression as implemented in scitkit-learn.

    Parameters
    -----------
    X : numpy.ndarray
        fit matrix
    y : numpy.ndarray
        target array
    fit_intercept : bool
        center data or not, forwarded to sklearn

    Returns
    ----------
    results : dict
        parameters
    """
    brr = BayesianRidge(fit_intercept=fit_intercept, **kwargs)
    brr.fit(X, y)
    results = dict()
    results['parameters'] = brr.coef_
    return results


def _fit_ardr(X, y, threshold_lambda=1e6, fit_intercept=False, **kwargs):
    """
    Returns the solution `a` to the linear problem `Xa=y` obtained by using
    the automatic relevance determination regression (ARDR) method as
    implemented in scitkit-learn.

    Parameters
    -----------
    X : numpy.ndarray
        fit matrix
    y : numpy.ndarray
        target array
    threshold_lambda : float
        threshold lambda parameter forwarded to sklearn
    fit_intercept : bool
        center data or not, forwarded to sklearn

    Returns
    ----------
    results : dict
        parameters
    """
    ardr = ARDRegression(threshold_lambda=threshold_lambda,
                         fit_intercept=fit_intercept, **kwargs)
    ardr.fit(X, y)
    results = dict()
    results['parameters'] = ardr.coef_
    return results


def _fit_rfe_l2(X, y, n_features=None, step=None, **kwargs):
    """Recursive feature elimination (RFE) L2 fitting

    RFE - L2 fitting is a method which runs recusrive feature elimination
    (as implemented in scikit-learn) with least-square fitting. The final model
    is obtained via a least-square fit using the selected features.

    Parameters
    -----------
    X : numpy.ndarray
        fit matrix
    y : numpy.ndarray
        target array
    n_features : int
        number of features to select, if None RFECV will be used to determine
        the optimal number of features
    step : int
        number of parameters to eliminate in each iteration

    Returns
    ----------
    results : dict
        parameters and selected features
    """

    n_params = X.shape[1]
    if step is None:
        step = int(np.ceil(n_params / 25))

    if n_features is None:
        return _fit_rfe_l2_CV(X, y, step, **kwargs)
    else:
        # extract features
        lr = LinearRegression(fit_intercept=False)
        rfe = RFE(lr, n_features_to_select=n_features, step=step, **kwargs)
        rfe.fit(X, y)
        features = rfe.support_

        # carry out final fit
        params = np.zeros(n_params)
        params[features] = _fit_least_squares(X[:, features], y)['parameters']

        # finish up
        results = dict(parameters=params, features=features)
        return results


def _fit_rfe_l2_CV(X, y, step=None, rank=1, n_jobs=-1, **kwargs):
    """Recursive feature elimination (RFE) L2 fitting with cross-validation (CV).

    Recursive feature elimination with least-squares fitting using
    cross-validation for optimizing the number of features.

    Parameters
    -----------
    X : numpy.ndarray
        fit matrix
    y : numpy.ndarray
        target array
    step : int
        number of parameters to eliminate in each iteration
    rank : int
        rank to use when selecting features

    Returns
    ----------
    results : dict
        parameters and selected features

    """

    n_params = X.shape[1]
    if step is None:
        step = int(np.ceil(n_params / 25))

    # setup
    cv = ShuffleSplit(train_size=0.9, test_size=0.1, n_splits=5)
    lr = LinearRegression(fit_intercept=False)
    rfecv = RFECV(lr, step=step, cv=cv, n_jobs=n_jobs,
                  scoring='neg_mean_squared_error', **kwargs)

    # extract features
    rfecv.fit(X, y)
    ranking = rfecv.ranking_
    features = ranking <= rank

    # carry out final fit
    params = np.zeros(n_params)
    params[features] = _fit_least_squares(X[:, features], y)['parameters']

    # finish up
    results = dict(parameters=params, features=features, ranking=ranking)
    return results


fit_methods = OrderedDict([
    ('least-squares', _fit_least_squares),
    ('lasso', _fit_lasso),
    ('elasticnet', _fit_elasticnet),
    ('bayesian-ridge', _fit_bayesian_ridge),
    ('ardr', _fit_ardr),
    ('rfe-l2', _fit_rfe_l2),
    ('split-bregman', fit_split_bregman)
    ])
available_fit_methods = list(fit_methods.keys())



"""
Optimizer with cross validation score
"""

import numpy as np
from sklearn.model_selection import KFold, ShuffleSplit
from .base_optimizer import BaseOptimizer
from .optimizer import Optimizer
from .fit_methods import fit
from .tools import ScatterData


validation_methods = {
    'k-fold': KFold,
    'shuffle-split': ShuffleSplit,
}


class CrossValidationEstimator(BaseOptimizer):
    """
    Optimizer with cross validation.

    This optimizer can compute cross-validation (CV) scores in
    different ways.  It can also produce the finalized model (using
    the full input data) for which the CV score is an estimation of
    its performance.

    Warning
    -------
    Repeatedly setting up a CrossValidationEstimator and training
    *without* changing the seed for the random number generator will yield
    identical or correlated results, to avoid this please specify a different
    seed when setting up multiple CrossValidationEstimator instances.

    Parameters
    ----------
    fit_data : tupe(numpy.ndarray, numpy.ndarray)
        the first element of the tuple represents the fit matrix `A`
        (`N, M` array) while the second element represents the vector
        of target values `y` (`N` array); here `N` (=rows of `A`,
        elements of `y`) equals the number of target values and `M`
        (=columns of `A`) equals the number of parameters
    fit_method : str
        method to be used for training; possible choice are
        "least-squares", "lasso", "elasticnet", "bayesian-ridge", "ardr",
        "rfe-l2", "split-bregman"
    standardize : bool
        whether or not to standardize the fit matrix before fitting
    validation_method : str
        method to use for cross-validation; possible choices are
        "shuffle-split", "k-fold"
    n_splits : int
        number of times the fit data set will be split for the cross-validation
    check_condition : bool
        whether or not to carry out a check of the condition number

        N.B.: This can be sligthly more time consuming for larger
        matrices.
    seed : int
        seed for pseudo random number generator

    Attributes
    ----------
    train_scatter_data : ScatterData
        contains target and predicted values from each individual
        traininig set in the cross-validation split;
        :class:`ScatterData` is a namedtuple.
    validation_scatter_data : ScatterData
        contains target and predicted values from each individual
        validation set in the cross-validation split;
        :class:`ScatterData` is a namedtuple.

    """

    def __init__(self, fit_data, fit_method='least-squares', standardize=True,
                 validation_method='k-fold', n_splits=10, check_condition=True,
                 seed=42, **kwargs):

        super().__init__(fit_data, fit_method, standardize, check_condition,
                         seed)

        if validation_method not in validation_methods.keys():
            msg = ['Validation method not available']
            msg += ['Please choose one of the following:']
            for key in validation_methods:
                msg += [' * ' + key]
            raise ValueError('\n'.join(msg))
        self._validation_method = validation_method
        self._n_splits = n_splits

        self._set_kwargs(kwargs)

        # data set splitting object
        self._splitter = validation_methods[validation_method](
            n_splits=self.n_splits, random_state=seed,
            **self._split_kwargs)

        self.train_scatter_data = None
        self.validation_scatter_data = None

        self._parameters_splits = None
        self._rmse_train_splits = None
        self._rmse_valid_splits = None
        self._rmse_train_final = None

    def train(self):
        """ Constructs the final model using all input data available. """
        self._fit_results = fit(self._A, self._y, self.fit_method,
                                self.standardize, self._check_condition,
                                **self._fit_kwargs)
        self._rmse_train_final = self.compute_rmse(self._A, self._y)

    def validate(self):
        """ Runs validation. """
        train_target, train_predicted = [], []
        valid_target, valid_predicted = [], []
        rmse_train_splits, rmse_valid_splits = [], []
        parameters_splits = []
        for train_set, test_set in self._splitter.split(self._A):
            opt = Optimizer((self._A, self._y), self.fit_method,
                            train_set=train_set,
                            test_set=test_set,
                            check_condition=self._check_condition,
                            **self._fit_kwargs)
            opt.train()

            parameters_splits.append(opt.parameters)
            rmse_train_splits.append(opt.rmse_train)
            rmse_valid_splits.append(opt.rmse_test)
            train_target.extend(opt.train_scatter_data.target)
            train_predicted.extend(opt.train_scatter_data.predicted)
            valid_target.extend(opt.test_scatter_data.target)
            valid_predicted.extend(opt.test_scatter_data.predicted)

        self._parameters_splits = np.array(parameters_splits)
        self._rmse_train_splits = np.array(rmse_train_splits)
        self._rmse_valid_splits = np.array(rmse_valid_splits)
        self.train_scatter_data = ScatterData(
            target=np.array(train_target), predicted=np.array(train_predicted))
        self.validation_scatter_data = ScatterData(
            target=np.array(valid_target), predicted=np.array(valid_predicted))

    def _set_kwargs(self, kwargs):
        """ Sets up fit_kwargs and split_kwargs.
        Different split methods need different keywords.
        """
        self._fit_kwargs = {}
        self._split_kwargs = {}

        if self.validation_method == 'k-fold':
            for key, val in kwargs.items():
                if key in ['shuffle']:
                    self._split_kwargs[key] = val
                else:
                    self._fit_kwargs[key] = val
        elif self.validation_method == 'shuffle-split':
            for key, val in kwargs.items():
                if key in ['test_size', 'train_size']:
                    self._split_kwargs[key] = val
                else:
                    self._fit_kwargs[key] = val

    @property
    def summary(self):
        """ dict : comprehensive information about the optimizer """
        info = super().summary

        # Add class specific data
        info['validation_method'] = self.validation_method
        info['n_splits'] = self.n_splits
        info['rmse_train_final'] = self.rmse_train_final
        info['rmse_train'] = self.rmse_train
        info['rmse_train_splits'] = self.rmse_train_splits
        info['rmse_validation'] = self.rmse_validation
        info['rmse_validation_splits'] = self.rmse_validation_splits
        info['train_scatter_data'] = self.train_scatter_data
        info['validation_scatter_data'] = self.validation_scatter_data

        # add kwargs used for fitting and splitting
        info = {**info, **self._fit_kwargs, **self._split_kwargs}
        return info

    def __repr__(self):
        kwargs = dict()
        kwargs['fit_method'] = self.fit_method
        kwargs['validation_method'] = self.validation_method
        kwargs['n_splits'] = self.n_splits
        kwargs['seed'] = self.seed
        kwargs = {**kwargs, **self._fit_kwargs, **self._split_kwargs}
        return 'CrossValidationEstimator((A, y), {})'.format(
            ', '.join('{}={}'.format(*kwarg) for kwarg in kwargs.items()))

    @property
    def validation_method(self):
        """ str : validation method name """
        return self._validation_method

    @property
    def n_splits(self):
        """ str : number of splits (folds) used for cross-validation """
        return self._n_splits

    @property
    def parameters_splits(self):
        """ numpy.ndarray : all parameters obtained during cross-validation """
        return self._parameters_splits

    @property
    def rmse_train_final(self):
        """
        float : root mean squared error when using the full set of input data
        """
        return self._rmse_train_final

    @property
    def rmse_train(self):
        """ float : average root mean squared training error obtained during
        cross-validation """
        if self._rmse_train_splits is None:
            return None
        return np.sqrt(np.mean(self._rmse_train_splits**2))

    @property
    def rmse_train_splits(self):
        """ list(float) : root mean squared training errors obtained during
        cross-validation """
        return self._rmse_train_splits

    @property
    def rmse_validation(self):
        """ float : average root mean squared cross-validation error """
        if self._rmse_valid_splits is None:
            return None
        return np.sqrt(np.mean(self._rmse_valid_splits**2))

    @property
    def rmse_validation_splits(self):
        """ list(float) : root mean squared validation errors obtained during
        cross-validation """
        return self._rmse_valid_splits



"""
This module contains various support/utility functions.
"""
import numpy as np
from ase.geometry import find_mic
from .io.logging import logger


logger = logger.getChild('utilities')


def get_displacements(atoms, atoms_ideal):
    """Computes the smallest possible displacements from positions and
    ideal positions given a cell metric.

    Notes
    -----
    * uses :func:`ase.geometry.find_mic`
    * assumes `pbc=[True, True, True]`.

    Parameters
    ----------
    positions : numpy.ndarray
        positions
    ideal_positions : numpy.ndarray
        ideal positions from which displacements are computed
    cell : numpy.ndarray
        cell metric with shape `(3,3)`

    Returns
    -------
    numpy.ndarray
        wrapped displacements
    """
    if not np.array_equal(atoms.numbers, atoms_ideal.numbers):
        raise ValueError('Atomic numbers do not match.')

    displacements = []
    for pos, ideal_pos in zip(atoms.positions, atoms_ideal.positions):
        v_ij = np.array([pos - ideal_pos])
        displacements.append(find_mic(v_ij, atoms_ideal.cell, pbc=True)[0][0])
    return np.array(displacements)


def prepare_structures(structures, atoms_ideal, calc):
    """
    Prepares a set of structures in the format suitable for a
    :class:`StructureContainer <hiphive.StructureContainer>`.

    **Note:** Changes the structures in place.
    """
    for atoms in structures:
        atoms.set_calculator(calc)
        forces = atoms.get_forces()
        displacements = get_displacements(atoms, atoms_ideal)
        atoms.positions = atoms_ideal.get_positions()
        atoms.new_array('forces', forces)
        atoms.new_array('displacements', displacements)


def get_neighbor_shells(atoms, cutoff, dist_tol=1e-5):
    """ Gets list of neighbor shells.

    Distances are grouped into shells via the following algorithm:

    1. Find smallest atomic distance `d_min`

    2. Find all pair distances in the range `d_min + 1 * dist_tol`

    3. Construct a shell from these and pop them from distance list

    4. Go to 1.

    Parameters
    ----------
    atoms : ase.Atoms
        Atoms used for finding shells
    cutoff : float
        exclude neighbor shells which have a distance larger than cutoff
    dist_tol : float
        distance tolerance

    Returns
    -------
    list(Shell)
        neighbor shells
    """

    # TODO: Remove this once this feature have been in ASE long enough
    try:
        from ase.neighborlist import neighbor_list
    except ImportError:
        raise ImportError('get_neighbor_shells uses new functionality from ASE'
                          ', please update ASE in order to use this function.')

    # get distances
    ijd = neighbor_list('ijd', atoms, cutoff)
    ijd = list(zip(*ijd))
    ijd.sort(key=lambda x: x[2])

    # sort into shells
    symbols = atoms.get_chemical_symbols()
    shells = []
    for i, j, d in ijd:
        types = tuple(sorted([symbols[i], symbols[j]]))
        for shell in shells:
            if abs(d - shell.distance) < dist_tol and types == shell.types:
                shell.count += 1
                break
        else:
            shell = Shell(types, d, 1)
            shells.append(shell)
    shells.sort(key=lambda x: (x.distance, x.types, x.count))

    # warning if two shells are within 2 * tol
    for i, s1 in enumerate(shells):
        for j, s2 in enumerate(shells[i+1:]):
            if s1.types != s2.types:
                continue
            if not s1.distance < s2.distance - 2 * dist_tol:
                logger.warning('Found two shells within 2 * dist_tol')

    return shells


class Shell:
    """
    Neighbor Shell class

    Parameters
    ----------
    types : list or tuple
        atomic types for neighbor shell
    distance : float
        interatomic distance for neighbor shell
    count : int
        number of pairs in the neighbor shell
    """

    def __init__(self, types, distance, count=0):
        self.types = types
        self.distance = distance
        self.count = count

    def __str__(self):
        s = '{}-{}   distance: {:10.6f}    count: {}'.format(
            *self.types, self.distance, self.count)
        return s

    __repr__ = __str__



"""
PhononHarmonics copied from ASE.
"""

import numpy as np
from ase.units import kB


def generate_phonon_rattled_structures(atoms, fc2, n_structures, temperature):
    """
    Returns list of phonon-rattled configurations.

    Configurations are generated by superimposing harmonic phonon
    eigenmodes with amplitudes corresponding to a certain temperatures
    and randomly drawn phase factors.

    Let :math:`\\boldsymbol{X}_{ai}` be the phonon modes indexed by atom
    :math:`a` and mode :math:`i`, :math:`\\omega_i` the phonon
    frequencies, and let :math:`0 < Q_i \\leq 1` and :math:`0 \\leq
    U_i < 1` be uniformly random numbers.  Then

    .. math::

        \\boldsymbol{R}_a
        = \\boldsymbol{R}^0_a
        + \\left<\\frac{k_B T}{m_a} \\right>^{1/2}
        \\sum_i \\frac{1}{\\omega_i} \\boldsymbol{X}_{ai}
        \\sqrt{-2 \\ln Q_i} \\cos(\\pi \\omega_i U_i)

    Parameters
    ----------
    atoms : ase.Atoms
        prototype structure
    fc2 : numpy.ndarray
        second order force constant matrix, with shape `(3N, 3N)`
    n_structures : int
        number of structures to generate
    temperature : float
        temperature in Kelvin

    Returns
    -------
    list(ase.Atoms)
        generated structures
    """
    if fc2.shape != (3 * len(atoms), 3 * len(atoms)):
        raise ValueError
    structures = []
    for _ in range(n_structures):
        atoms_tmp = atoms.copy()
        PhononHarmonics(atoms_tmp, temperature*kB, fc2)
        structures.append(atoms_tmp)
    return structures


def PhononHarmonics(atoms, temp, force_constants, rng=np.random,
                    failfast=True):
    """
    Excites phonon modes to specified temperature.

    This excites all phonon modes randomly so that each contributes,
    on average, equally to the given temperature.  Both potential
    energy and kinetic energy will be consistent with the phononic
    vibrations characteristic of the specified temperature.

    Parameters
    ----------
    atoms : ase.Atoms
        prototype structure
    temp : float
        temperature in eV
    force_constants : numpy.ndarray
        second order force constant matrix, with shape `(3N, 3N)`
    rng : func
        random number generation
    failfast : True

    Raises
    ------
    ValueError
        if egenvalues of the dynamical matrix look suspicious (they
        should be three zeros for translational modes followed by any
        number of strictly positive eigenvalues).
    """
    mass_a = atoms.get_masses()

    rminv = (mass_a**-0.5).repeat(3)
    dynamical_matrix = force_constants.copy()
    dynamical_matrix *= rminv[:, None]
    dynamical_matrix *= rminv[None, :]

    w2_s, X_is = np.linalg.eigh(dynamical_matrix)

    # set imaginary frequencies to positive
    w2_s = np.sort(np.abs(w2_s))

    if failfast:
        zeros = w2_s[:3]
        worst_zero = np.abs(zeros).max()
        if worst_zero > 1e-3:
            raise ValueError('Translational modes have suspiciously large '
                             'energies; should be close to zero: {}'
                             .format(w2_s[:3]))

        w2min = w2_s[3:].min()
        if w2min < 0:
            raise ValueError('Dynamical matrix has negative eigenvalues '
                             'such as {}'.format(w2min))

    # First three modes are translational so ignore:
    nw = len(w2_s) - 3
    w_s = np.sqrt(w2_s[3:])
    X_acs = X_is[:, 3:].reshape(len(atoms), 3, nw)

    # We need 0 < P <= 1.0 and not 0 0 <= P < 1.0 for the logarithm
    # to avoid (highly improbable) NaN:
    A_s = np.sqrt(-2.0 * np.log(1.0 - rng.rand(nw)))
    phi_s = 2.0 * np.pi * rng.rand(nw)

    d_ac = (A_s * np.sin(phi_s) / w_s * X_acs).sum(axis=2)
    v_ac = (A_s * np.cos(phi_s) * X_acs).sum(axis=2)

    sqrtkTm_a = np.sqrt(temp / mass_a)
    for quantity in [d_ac, v_ac]:
        quantity *= sqrtkTm_a[:, None]

    atoms.positions += d_ac
    atoms.set_velocities(v_ac)



# Contains the get_clusters function which generates clusters

from ase.neighborlist import NeighborList
import numpy as np
import itertools as it
from collections import defaultdict

from .utilities import BiMap
from ..io.logging import logger

logger = logger.getChild('get_clusters')


# TODO: This function could be made a bit more general
def get_clusters(atoms, cutoffs, nPrim, multiplicity=True,
                 use_geometrical_order=False):
    """Generate a list of all clusters in the atoms object which includes the
    center atoms with positions within the cell metric. The cutoff determines
    up to which order and range clusters should be generated.

    With multiplicity set to True clusters like `[0,0]` and `[3,3,4` etc will
    be generated. This is useful when doing force constants but not so much for
    cluster expansions.

    The geometrical order is the total number of different atoms in the
    cluster. `[0,0,1]` would have geometrical order 2 and `[1,2,3,4]` would
    have order 4. If the key word is True the cutoff criteria will be based on
    the geometrical order of the cluster. This is based on the observation that
    many body interactions decrease fast with cutoff but anharmonic
    interactions can be quite long ranged.

    Parameters
    ----------
    atoms : ase.Atoms
        can be a general atoms object but must have pbc=False.
    cutoffs : dict
        the keys specify the order while the values specify the cutoff radii
    multiplicity : bool
        includes clusters where same atom appears more than once
    geometrical_order : bool
        specifies if the geometrical order should be used as cutoff_order,
        otherwise the normal order of the cluster is used

    Returns
    -------
    list(tuple(int))
        a list of clusters where each entry is a tuple of indices,
        which refer to the atoms in the input supercell
    """

    logger.debug('Generating clusters...')
    cluster_dict = defaultdict(list)
    for i in range(nPrim):
        for order in cutoffs.orders:
            cluster = (i,) * order
            cluster_dict[order].append(cluster)
    for nbody in range(2, cutoffs.max_nbody + 1):
        cutoff = cutoffs.max_nbody_cutoff(nbody)
        nbody_clusters, nbody_cutoffs = \
            generate_geometrical_clusters(atoms, nPrim, cutoff, nbody)
        for order in range(nbody, cutoffs.max_nbody_order(nbody) + 1):
            for cluster, cutoff in zip(nbody_clusters, nbody_cutoffs):
                if cutoff < cutoffs.get_cutoff(nbody=nbody, order=order):
                    clusters = extend_cluster(cluster, order)
                    cluster_dict[order].extend(clusters)
    cluster_list = BiMap()
    for key in sorted(cluster_dict):
        for cluster in sorted(cluster_dict[key]):
            cluster_list.append(cluster)
    return cluster_list


def generate_geometrical_clusters(atoms, nPrim, cutoff, order):
    nm, dm = create_neighbor_matrices(atoms, cutoff)
    clusters = []
    cutoffs = []
    i, j = 0, 0
    for tup in it.combinations(range(len(atoms)), r=order):
        if tup[0] >= nPrim:
            break
        if not nm[tup[i], tup[j]]:
            continue
        for i, j in it.combinations(range(order), r=2):
            if not nm[tup[i], tup[j]]:
                break
        else:
            clusters.append(tup)
            cutoffs.append(np.max(dm[tup, :][:, tup]))
    return clusters, cutoffs


def create_neighbor_matrices(atoms, cutoff):
    atoms = atoms.copy()
    atoms.pbc = False
    nAtoms = len(atoms)
    nl = NeighborList([cutoff / 2] * nAtoms, skin=0, bothways=True)
    nl.update(atoms)
    neighbor_matrix = np.eye(nAtoms, dtype=bool)
    distance_matrix = np.zeros((nAtoms, nAtoms))
    for i in range(nAtoms):
        indices, _ = nl.get_neighbors(i)
        neighbor_matrix[i, indices] = True
        distance_matrix[i, indices] = atoms.get_distances(i, indices)
    return neighbor_matrix, distance_matrix


def extend_cluster(cluster, order):
    clusters = []
    cluster = tuple(cluster)
    nbody = len(cluster)
    r = order - nbody
    for tup in it.combinations_with_replacement(cluster, r):
        new_cluster = sorted(cluster + tup)
        clusters.append(tuple(new_cluster))
    return clusters



"""
Contains the force constant modell (FCM) which handles cluster and force
constant information of a super cell object.
"""
import math
import pickle
import numpy as np

from .cluster_space import ClusterSpace
from .core.atoms import Atom, atom_to_spos, spos_to_atom, align_super_cell
from .core.orbit import Orbit
from .core.orientation_family import OrientationFamily
from .core.tensors import rotation_to_cart_coord, rotate_tensor
from .core.utilities import Progress, BiMap
from .force_constants import ForceConstants
from .io.logging import logger
from .calculators.numba_calc import (clusters_force_contribution,
                                     cluster_force_contribution)

logger = logger.getChild('fcm')


class ForceConstantModel:
    """Transfers a cluster space onto a super structure

    Contains the full description of all clusters and the force constants
    within a super cell with periodic boundary conditions.

    Parameters
    ----------
    atoms : ase.Atoms
        configuration to which the cluster space is to be applied
    cs : ClusterSpace
        a cluster space compatible with the structure of the atoms
    """
    def __init__(self, atoms, cs):

        self.atoms = atoms.copy()
        self.orbits = []
        self.cluster_list = BiMap()
        self.cs = cs
        self._populate_orbits(atoms)

    # TODO: refactor
    def _populate_orbits(self, atoms):
        """Map the orbits from the underlying force constant potential onto the
        supercell structure associated with this force constant model.

        """
        # TODO: Comment function
        atom_lookup = {}
        cs = self.cs

        aligned_super_cell, scR, _ = align_super_cell(
            atoms, cs.primitive_structure, cs.symprec)
        sc = aligned_super_cell.copy()
        sc.pbc = False
        sc.cell = cs.primitive_structure.cell
        atom_list = BiMap()
        # TODO: fix this. see also 334d8572
        for spos in sc.get_scaled_positions():
            atom_list.append(spos_to_atom(spos, cs.primitive_structure.basis))
        sorted_cluster_list = BiMap()
        if hasattr(cs, 'rotation_matrices'):
            rotations = []
            for R in cs.rotation_matrices:
                rotations.append(rotation_to_cart_coord(
                    R, cs.primitive_structure.cell))

        def get_atom_index(atom):
            tupd_atom = (atom.site, *atom.offset)
            if tupd_atom in atom_lookup:
                return atom_lookup[tupd_atom]
            spos = atom_to_spos(atom, cs.primitive_structure.basis)
            pos = np.dot(spos, cs.primitive_structure.cell)
            spos = np.dot(pos, np.linalg.inv(aligned_super_cell.cell))
            atom = spos_to_atom(spos, aligned_super_cell.basis)
            atom_lookup[tupd_atom] = atom.site
            return atom.site

        def get_mapped_cluster(cluster, offset):
            new_cluster = []
            for atom_index in cluster:
                atom = cs.atom_list[atom_index]
                translated_atom = Atom(atom.site, np.add(atom.offset, offset))
                index = get_atom_index(translated_atom)
                new_cluster.append(index)
            return new_cluster
        logger.info('Populating orbits')
        bar = Progress(len(cs.orbits))
        for orbit_index, orbit in enumerate(cs.orbits):

            new_orbit = Orbit()
            new_orbit.order = orbit.order

            if len(orbit.eigentensors) > 0:
                ets = []
                for et in orbit.eigentensors:
                    ets.append(rotate_tensor(et, scR))
                new_orbit.eigentensors = ets
                new_orbit.force_constant = np.zeros(ets[0].shape)
            else:
                new_orbit.force_constant = rotate_tensor(orbit.force_constant,
                                                         scR)
            cluster = cs.cluster_list[orbit.prototype_index]
            _, pos, counts = np.unique(np.array(cluster),
                                       return_index=True, return_counts=True)
            new_orbit.positions = pos
            prefactor = -1 / np.prod(list(map(math.factorial, counts)))
            new_orbit.prefactors = np.array([prefactor * c for c in counts])

            for of in orbit.orientation_families:

                new_of = OrientationFamily()
                if len(orbit.eigentensors) > 0:
                    ets = []
                    R_inv = rotations[of.symmetry_index].T
                    for et in orbit.eigentensors:
                        et_of = rotate_tensor(et, R_inv)
                        ets.append(rotate_tensor(et_of, scR))
                    new_of.eigentensors = ets
                    new_of.force_constant = np.zeros(ets[0].shape)
                else:
                    new_of.force_constant = rotate_tensor(of.force_constant,
                                                          scR)

                cluster = cs.cluster_list[of.cluster_indices[0]]
                if isinstance(cs, ClusterSpace):
                    perm = cs._permutations[of.permutation_indices[0]]
                    cluster = [cluster[i] for i in np.argsort(perm)]
                for atom in atom_list:
                    if not atom.site == cs.atom_list[cluster[0]].site:
                        continue
                    offset = atom.offset
                    new_cluster = tuple(get_mapped_cluster(cluster, offset))
                    sorted_new_cluster = tuple(sorted(new_cluster))
                    if sorted_new_cluster in sorted_cluster_list:
                        raise Exception('Found cluster {} twice, check '
                                        'cutoff!'.format(sorted_new_cluster))
                    sorted_cluster_list.append(sorted_new_cluster)
                    self.cluster_list.append(new_cluster)
                    new_cluster_index = len(self.cluster_list) - 1
                    new_of.cluster_indices.append(new_cluster_index)
                new_orbit.orientation_families.append(new_of)
            self.orbits.append(new_orbit)
            bar.tick()
        bar.close()
        if isinstance(cs, ClusterSpace):
            self._parameters = np.zeros(self.cs.n_dofs)

    def get_force_constants(self):
        """Obtain the force constants of the super cell

        Returns
        -------
        force constants : ForceConstants
            the complete set of force constants
        """

        fcs = []
        clusters = []
        for orbit in self.orbits:
            for of in orbit.orientation_families:
                fc = of.force_constant
                fcs.append(fc.copy())
                of_clusters = []
                for cluster_index in of.cluster_indices:
                    cluster = self.cluster_list[cluster_index]
                    of_clusters.append(tuple(cluster))
                clusters.append(of_clusters)
        return ForceConstants(
            cluster_groups=clusters, fc_list=fcs, atoms=self.atoms)

    @property
    def parameters(self):
        """ list(float): parameters """
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        self._parameters = parameters
        mapped_parameters = self.cs._map_parameters(parameters)
        p = 0
        for orb in self.orbits:
            fc_is_zero = np.allclose(orb.force_constant, 0)
            params_is_zero = np.allclose(
                mapped_parameters[p: p+len(orb.eigentensors)], 0)
            if fc_is_zero and params_is_zero:
                p += len(orb.eigentensors)
                continue
            orb.force_constant *= 0
            if not params_is_zero:
                for et, a in zip(orb.eigentensors, mapped_parameters[p:]):
                    orb.force_constant += et * a
            for of in orb.orientation_families:
                of.force_constant *= 0
                if not params_is_zero:
                    for et, a in zip(of.eigentensors, mapped_parameters[p:]):
                        of.force_constant += et * a
            p += len(orb.eigentensors)

    def get_forces(self, displacements):
        """ Returns the forces in the system given displacements.

        The parameters of the model must be set to get any result.

        Parameters
        ----------
        displacements : numpy.ndarray
            displacements of each atom in the supercell (`N, 3` array)
        """
        F = np.zeros(displacements.shape)
        f = np.zeros(3)
        for orbit in self.orbits:
            if np.allclose(orbit.force_constant, 0):
                continue
            order = orbit.order
            positions = orbit.positions
            prefactors = orbit.prefactors
            for of in orbit.orientation_families:
                fc = of.force_constant.flatten()
                fc_tmp = fc.copy()

                for cluster_index in of.cluster_indices:
                    cluster = self.cluster_list[cluster_index]
                    cluster_force_contribution(
                            positions, prefactors, len(prefactors),
                            fc_tmp, fc, order,
                            displacements,
                            cluster, f, F)
        return F

    def get_fit_matrix(self, displacements):
        """ Returns the matrix used to fit the parameters.

        Represents the linear relation between the parameters and the forces.

        Parameters
        ----------
        displacements : numpy.ndarrya
            displacements of each atom in the supercell (`(N, 3)` array)
        """

        F = np.zeros(displacements.shape)
        f = np.zeros(3)
        M = np.zeros((F.size, self.cs._cvs.shape[0]))
        et_index = 0
        bar = Progress(len(self.orbits))
        for orbit in self.orbits:
            bar.tick()
            fc_tmp = np.zeros(3**orbit.order)
            for of in orbit.orientation_families:
                clusters = np.zeros((len(of.cluster_indices), orbit.order),
                                    dtype=np.int64)
                for i, cluster_index in enumerate(of.cluster_indices):
                    clusters[i] = self.cluster_list[cluster_index]
                for i, et in enumerate(of.eigentensors):
                    F[:] = 0
                    clusters_force_contribution(orbit.positions,
                                                orbit.prefactors,
                                                len(orbit.prefactors),
                                                fc_tmp,
                                                et.ravel(),
                                                orbit.order,
                                                displacements,
                                                clusters, f, F)
                    M[:, et_index + i] += F.flat
            et_index += len(orbit.eigentensors)
        M = M.dot(self.cs._cvs.toarray())
        bar.close()
        return M

    @staticmethod
    def read(f):
        """Reads a force constant model from file.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to load from (file object)

        Returns
        -------
        ForceConstantModel
            the original object as stored in the file
        """
        if isinstance(f, str):
            with open(f, 'rb') as fobj:
                return pickle.load(fobj)
        else:
            return pickle.load(f)

    def write(self, f):
        """Writes a force constant model to file.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to write to (file object)
        """
        if isinstance(f, str):
            with open(f, 'wb') as fobj:
                pickle.dump(self, fobj)
        else:
            pickle.dump(self, f)



def read_force_constants(filename, format=None):
    """
    Reads force constants from file. Several different file formats
    are supported including `phonopy`, `phono3py`, and `shengBTE`.

    Parameters
    ----------
    filename : str
        name of input file
    format : str
        format of input file

        possible values:
        ``phonopy``, ``phonopy-text``, ``phonopy-hdf5``, ``phono3py``,
        ``shengbte``

        In the case of ``phonopy`` hiphive will attempt to infer the file
        format (hdf5 or text) from the file name.

    Returns
    -------
    numpy.ndarray
        second (`phonopy`) or third-order (`phono3py`, `shengBTE`)
        force constant matrix

    Raises
    ------
    ValueError
        if file format cannot be recognized
    """
    from .phonopy import read_phonopy_fc2, read_phonopy_fc3
    from .shengBTE import read_shengBTE_fc3

    if format == 'phonopy':
        return read_phonopy_fc2(filename)
    elif format == 'phonopy-text':
        return read_phonopy_fc2(filename, format='text')
    elif format == 'phonopy-hdf5':
        return read_phonopy_fc2(filename, format='hdf5')
    elif format == 'phono3py':
        return read_phonopy_fc3(filename)
    elif format == 'shengbte':
        return read_shengBTE_fc3(filename)
    else:
        raise ValueError('Unrecognized file format: {}'.format(format))


def write_force_constants(filename, fc, format=None):
    """
    Writes force constants to file.

    Parameters
    ----------
    filename : str
        name of output file
    fc : ForceConstants or numpy.ndarray
        force constant matrix
    format : str
        format of output file

        possible values:
        ``phonopy``, ``phonopy-text``, ``phonopy-hdf5``, ``phono3py``,
        ``shengbte``

        In the case of `phonopy` hiphive will attempt to infer the file
        format (hdf5 or text) from the file name.

    Raises
    ------
    ValueError
        if file format cannot be recognized
    """
    from .phonopy import write_phonopy_fc2, write_phonopy_fc3
    from .shengBTE import write_shengBTE_fc3

    if format == 'phonopy':
        write_phonopy_fc2(filename, fc)
    elif format == 'phonopy-text':
        write_phonopy_fc2(filename, fc, format='text')
    elif format == 'phonopy-hdf5':
        write_phonopy_fc2(filename, fc, format='hdf5')
    elif format == 'phono3py':
        write_phonopy_fc3(filename, fc)
    elif format == 'shengbte':
        write_shengBTE_fc3(filename, fc)
    else:
        raise ValueError('Unrecognized file format: {}'.format(format))



"""
This module provides functionality for storing and handling of force constants
"""
import tarfile
import numpy as np

from ase import units
from itertools import permutations, product
from scipy.linalg import eig
from string import ascii_lowercase, ascii_uppercase

from .io.read_write_files import add_items_to_tarfile_pickle, read_items_pickle


class ForceConstants:
    """Container class for force constants.

    Either specify `fc_dict` or both `cluster_groups` and `fc_list`.

    Parameters
    ----------
    fc_dict : dict
        dict which holds all force constants with clusters as keys and the
        respective force constant as value
    cluster_groups : list
        list of groups of clusters, clusters in the same group should have
        identical force constants.
    fc_list : list
        list of force constants, one force constant for each cluster group
    atoms: ase.Atoms
        supercell corresponding to the fcs

    Attributes
    ----------
    _fc_dict : dict
        dictionary that holds all force constants with clusters as keys and the
        respective force constant as value
    """

    def __init__(self, fc_dict=None, cluster_groups=None, fc_list=None,
                 atoms=None):

        # set up fc_dict
        if fc_dict is None:
            self._set_up_fc_dict(cluster_groups, fc_list)
            self._sparse = True
            self._cluster_groups = cluster_groups
            self._fc_list = fc_list
        else:
            self._fc_dict = fc_dict
            self._sparse = False

        # find which orders are in the fc_dict
        orders = np.unique([len(c) for c in self._fc_dict.keys()])
        self._orders = orders.tolist()

        # set up atoms and allowed_atoms
        self._set_up_atoms(atoms)

    def _set_up_fc_dict(self, cluster_groups, fc_list):
        """ Set up fc_dict from cluster_groups and fc_list """
        if cluster_groups is None or fc_list is None:
            raise ValueError('Invalid input')
        if len(cluster_groups) != len(fc_list):
            raise ValueError('cluster_groups and fc_list not of same length')

        fc_dict = {}
        for clusters, fc in zip(cluster_groups, fc_list):
            for cluster in clusters:
                permutation = np.argsort(cluster)
                cluster = tuple(sorted(cluster))
                if cluster in fc_dict.keys():
                    raise ValueError('cluster_groups contains duplicates')
                fc_dict[cluster] = fc.transpose(permutation)

        self._fc_dict = fc_dict

    def _set_up_atoms(self, atoms):
        """ Set up atoms and sanity check its properties with fc_dict """
        cluster_indices = [ind for c in self._fc_dict.keys() for ind in c]
        natoms = np.max(cluster_indices) + 1
        if atoms is None:
            self._atoms = None
        else:
            self._atoms = atoms.copy()
            if natoms != len(self.atoms):
                raise ValueError('Atoms and ForceConstants have different '
                                 'number of atoms')

        self._allowed_atoms = set(range(natoms))
        if len(self._allowed_atoms - set(cluster_indices)) != 0:
            raise ValueError('Not all atoms are part of a force constant')

    def get_fc_array(self, order, format='phonopy'):
        """ Returns force constants in array format for specified order.

        Parameters
        ----------
        order : int
            force constants for this order will be returned
        format : str
            specify which format (shape) the NumPy array should have,
            possible values are `phonopy` and `ase`

        Returns
        -------
        numpy.ndarray
            NumPy array with shape `(N,)*order + (3,)*order` where `N` is
            the number of atoms
        """
        if order not in self.orders:
            raise ValueError('Order not in orders')
        if format not in ['ase', 'phonopy']:
            raise ValueError('Format must be either ase or phonopy')

        if format == 'ase':
            if order != 2:
                raise ValueError('ASE format works only for order 2')
            fc_array = self.get_fc_array(order, format='phonopy')
            return fc_array.transpose([0, 2, 1, 3]).reshape(
                self.natoms * 3, self.natoms * 3)

        fc_array = np.zeros((self.natoms,)*order + (3,)*order)
        for c, fc in self.get_fc_dict(order=order, permutations=True).items():
            fc_array[c] = fc
        return fc_array

    def get_fc_dict(self, order=None, permutations=False):
        """ Returns force constant dictionary for one specific order.

        Parameters
        ----------
        order : int
            fcs returned for this order
        permutations : bool
            if True returns all permutations of cluster, else only force
            constants for sorted cluster

        Returns
        -------
        dict
            dictionary with keys corresponding to clusters and values to the
            respective force constant
         """

        # Return all fcs
        if order is None:
            if not permutations:
                return self._fc_dict
            else:
                fc_dict = {}
                for c, fc in self._fc_dict.items():
                    for c_perm, fc_perm in get_permuted_fcs(c, fc).items():
                        fc_dict[c_perm] = fc_perm
                return fc_dict

        if order not in self.orders:
            raise ValueError('Order not in orders')

        # Return fcs for specific order
        fc_order = {}
        for c, fc in self._fc_dict.items():
            if len(c) == order:
                if permutations:
                    for c_perm, fc_perm in get_permuted_fcs(c, fc).items():
                        fc_order[c_perm] = fc_perm
                else:
                    fc_order[c] = fc
        return fc_order

    def __getitem__(self, cluster):
        if isinstance(cluster, (tuple, list)):
            if len(cluster) not in self.orders:
                raise ValueError('Cluster order not in orders')
            if len(self._allowed_atoms.intersection(set(cluster))) != \
                    len(set(cluster)):
                raise ValueError('Cluster outside range of atomic indices')

            try:
                sorted_cluster = tuple(sorted(cluster))
                if cluster == sorted_cluster:
                    return self._fc_dict[cluster]
                inv_perm = np.argsort(np.argsort(cluster))
                return self._fc_dict[sorted_cluster].transpose(inv_perm)
            except KeyError:
                return np.zeros((3,)*len(cluster))
        else:
            raise ValueError('Cluster must be tuple or list')

    def assert_acoustic_sum_rules(self, order=None, tol=1e-6):
        """ Asserts that acoustic sum rules are enforced for force constants.

        Parameters
        ----------
        order : int
            specifies which order to check, if None all are checked
        tol : float
            numeric tolerance for checking sum rules

        Raises
        ------
        AssertionError
            if acoustic sum rules are not enforced
        """

        # set up orders
        if order is None:
            orders = self.orders
        else:
            if order not in self.orders:
                raise ValueError('Order not available in FCS')
            orders = [order]

        atomic_indices = range(self.natoms)
        for order in orders:
            assert_msg = 'Acoustic sum rule order {} not enforced for atom' + \
                ' {}'*(order-1) + ' x'
            for ijk in product(atomic_indices, repeat=order-1):
                fc_sum_ijk = np.zeros((3, )*order)
                for l in atomic_indices:
                    cluster = ijk + (l, )
                    fc_sum_ijk += self[cluster]
                assert np.linalg.norm(fc_sum_ijk) < tol, \
                    assert_msg.format(order, *ijk)

    def __len__(self):
        return len(self._fc_dict)

    @property
    def atoms(self):
        """ ase.Atoms : supercell corresponding to force constants """
        return self._atoms

    @property
    def natoms(self):
        """ int : number of atoms (maximum index in a cluster +1) """
        return len(self._allowed_atoms)

    @property
    def orders(self):
        """ list : orders for which force constants exist """
        return self._orders

    @property
    def sparse(self):
        """ bool : if True the object was initialized with sparse data """
        return self._sparse

    @property
    def clusters(self):
        """ list : sorted list of clusters (identified as tuple of site
        indices) """
        return sorted(self._fc_dict.keys(), key=lambda c: (len(c), c))

    def write(self, f):
        """Writes force constants to file.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to write to (file object)
        """
        if isinstance(f, str):
            tar_file = tarfile.open(mode='w', name=f)
        else:
            tar_file = tarfile.open(mode='w', fileobj=f)

        # add the needed info to recreate the fcs object
        if self.sparse:
            items = dict(cluster_groups=self._cluster_groups,
                         fc_list=self._fc_list, atoms=self.atoms)
            add_items_to_tarfile_pickle(tar_file, items, 'repr')
        else:
            items = dict(fc_dict=self._fc_dict, atoms=self.atoms)
            add_items_to_tarfile_pickle(tar_file, items, 'repr')

        tar_file.close()

    @staticmethod
    def read(f):
        """Reads force constants from file.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to load from (file object)
        """
        if isinstance(f, str):
            tar_file = tarfile.open(mode='r', name=f)
        else:
            tar_file = tarfile.open(mode='r', fileobj=f)

        items = read_items_pickle(tar_file, 'repr')
        return ForceConstants(**items)

    def print_cluster(self, cluster: tuple):
        """
        Prints force constants for a cluster in a nice format.

        Parameters
        ----------
        cluster : tuple(int)
            sites belonging to the cluster
        """
        print(self._repr_fc(cluster))

    def __str__(self):
        s = []
        s.append(' ForceConstants '.center(54, '='))
        s.append('Orders: {}'.format(self.orders))
        s.append('Atoms: {}'.format(self.atoms))
        s.append('')
        if len(self) > 10:
            for c in self.clusters[:3]:
                s.append(self._repr_fc(c)+'\n')
            s.append('...\n')
            for c in self.clusters[-3:]:
                s.append(self._repr_fc(c)+'\n')
        else:
            for c in self.clusters:
                s.append(self._repr_fc(c)+'\n')
        return '\n'.join(s)

    def __repr__(self):
        fc_dict_str = '{{{}: {}, ..., {}: {}}}'.format(
            self.clusters[0], self[self.clusters[0]].round(5),
            self.clusters[-1], self[self.clusters[-1]].round(5))
        return 'ForceConstants(fc_dict={}, atoms={!r})'.format(
            fc_dict_str, self.atoms)

    def _repr_fc(self, cluster: tuple, precision=5, suppress=True) -> str:
        """
        Representation for single cluster and its force constant.

        Parameters
        ----------
        cluster : tuple
            tuple of ints indicating the sites belonging to the cluster
        """
        s = []
        s.append('Cluster: {}'.format(cluster))
        for atom_index in cluster:
            s.append(str(self.atoms[atom_index]))
        s.append('force constant:')
        s.append(np.array_str(self[cluster], precision=precision,
                 suppress_small=suppress))
        return '\n'.join(s)


def get_permuted_fcs(cluster, fc):
    """Generate all permutations of cluster and their force constant given a
    single cluster and its force constant.

    Parameters
    ----------
    cluster : tuple
        tuple of ints indicating the sites belonging to the cluster
    fc : numpy.ndarray
        force constant for cluster

    Returns
    -------
    dict
        dict with keys corresponding to clusters and values to their force
        constant
    """
    fc_dict = {}
    fc = fc.transpose(np.argsort(cluster))
    for perm_cluster in set(permutations(cluster)):
        inv_perm = np.argsort(np.argsort(perm_cluster))
        fc_perm = fc.transpose(inv_perm)
        fc_dict[perm_cluster] = fc_perm
    return fc_dict


def fc_array_to_dict(fc_array, cut_tol=1e-8):
    """Convert a force constant from array to dictionary format.

    Parameters
    ----------
    fc_array : numpy.ndarray
        force constant array for one order
    cut_tol : float
        tolerance for which force constants to keep in dict

    Returns
    -------
    dict
        sparse dictionary with keys corresponding to clusters and values to the
        respective force constant
    """

    fc_dict = {}
    Natoms = fc_array.shape[0]
    order = len(fc_array.shape) // 2
    if fc_array.shape != (Natoms,)*order + (3,)*order:
        raise ValueError('fc_array shape does not have expected shape')

    for cluster in product(*[range(Natoms)]*order):
        if np.linalg.norm(fc_array[cluster]) > cut_tol:
            if cluster == tuple(sorted(cluster)):
                fc_dict[cluster] = fc_array[cluster]
    return fc_dict


def compute_gamma_frequencies(fc2, masses):
    """ Computes gamma frequencies using second order force constants

    Assumes fc2 is in eV/A2

    Parameters
    ----------
    fc2 : ForceConstants or numpy.ndarray
        second order force constants
    masses : list(float)
        mass for each atom

    Returns
    -------
    gamma_frequencies : numpy.ndarray
        Gamma frequencies in THz
    """

    # get fc2_array
    if isinstance(fc2, ForceConstants):
        fc2 = fc2.get_fc_array(order=2)

    N = fc2.shape[0]
    if len(masses) != N:
        raise ValueError('Length of masses not compatible with fc2')
    mass_matrix = np.sqrt(np.outer(masses, masses))

    # divide with mass matrix
    fc2_tmp = np.zeros((N, N, 3, 3))
    for pair in product(range(N), repeat=2):
        fc2_tmp[pair] = fc2[pair] / mass_matrix[pair]

    # reshape into matrix and solve eigenvalues
    fc2_tmp = fc2_tmp.transpose([0, 2, 1, 3]).reshape(N * 3, N * 3)
    eigen_vals, _ = eig(fc2_tmp)
    eigen_vals *= 1e20 / units.J * units.kg  # [1/s**2]
    eigen_vals.sort()

    # if negative eigenval, set frequency to negative
    gamma_frequencies = []
    for val in eigen_vals.real:
        if val >= 0:
            gamma_frequencies.append(np.sqrt(val))
        else:
            gamma_frequencies.append(-np.sqrt(np.abs(val)))

    # Sort and convert to THz
    gamma_frequencies = np.array(gamma_frequencies) / 1e12 / (2 * np.pi)
    gamma_frequencies.sort()
    return gamma_frequencies


def symbolize_force_constant(fc, tol=1e-10):
    """Carry out a symbolic symmetrization of a force constant tensor.

    Parameters
    ----------
    fc : numpy.ndarray
        force constant tensor
    tol : float
        tolerance used to decide whether two elements are identical

    Returns
    -------
    numpy.ndarray
        symbolic representation of force constant matrix
    """

    fc_int = np.round(fc / tol).astype(int)

    fc_chars = np.empty(fc_int.shape, dtype=object).flatten()
    all_chars = ascii_lowercase + ascii_uppercase
    lookup_chars = {}
    for i, val in enumerate(fc_int.flatten()):
        if val == 0:
            fc_chars[i] = 0
        elif val in lookup_chars.keys():
            fc_chars[i] = lookup_chars[val]
        elif -val in lookup_chars.keys():
            fc_chars[i] = '-{:}'.format(lookup_chars[-val])
        else:
            lookup_chars[val] = all_chars[len(lookup_chars.keys())]
            fc_chars[i] = lookup_chars[val]
    return fc_chars.reshape(fc_int.shape)




"""
BaseOptimizer serves as base for all optimizers.
"""

import numpy as np
from abc import ABC, abstractmethod
from .fit_methods import available_fit_methods


class BaseOptimizer(ABC):
    """
    BaseOptimizer class.

    Serves as base class for all Optimizers solving `Ax = y`.

    Parameters
    ----------
    fit_data : tuple of NumPy (N, M) array and NumPy (N) array
        the first element of the tuple represents the fit matrix `A`
        whereas the second element represents the vector of target
        values `y`; here `N` (=rows of `A`, elements of `y`) equals the number
        of target values and `M` (=columns of `A`) equals the number of
        parameters
    fit_method : str
        method to be used for training; possible choice are
        "least-squares", "lasso", "elasticnet", "bayesian-ridge", "ardr",
        "rfe-l2", "split-bregman"
    standardize : bool
        whether or not to standardize the fit matrix before fitting
    check_condition : bool
        whether or not to carry out a check of the condition number

        N.B.: This can be sligthly more time consuming for larger
        matrices.
    seed : int
        seed for pseudo random number generator
    """

    def __init__(self, fit_data, fit_method, standardize=True,
                 check_condition=True, seed=42):
        """
        Attributes
        ----------
        _A : NumPy (N, M) array
            fit matrix
        _y : NumPy (N) array
            target values
        """

        if fit_method not in available_fit_methods:
            raise ValueError('Unknown fit_method: {}'.format(fit_method))

        if fit_data is None:
            raise TypeError('Invalid fit data; Fit data can not be None')
        if fit_data[0].shape[0] != fit_data[1].shape[0]:
            raise ValueError('Invalid fit data; shapes of fit matrix'
                             ' and target vector do not match')
        if len(fit_data[0].shape) != 2:
            raise ValueError('Invalid fit matrix; must have two dimensions')

        self._A, self._y = fit_data
        self._n_rows = self._A.shape[0]
        self._n_cols = self._A.shape[1]
        self._fit_method = fit_method
        self._standarize = standardize
        self._check_condition = check_condition
        self._seed = seed
        self._fit_results = {'parameters': None}

    def compute_rmse(self, A, y):
        """
        Computes the root mean square error using :math:`A`, :math:`y`,
        and the vector of fitted parameters :math:`x` corresponding to
        :math:`||Ax-y||_2`.

        Parameters
        ----------
        A : numpy.ndarray
            fit matrix (`N,M` array) where `N` (=rows of `A`, elements
            of `y`) equals the number of target values and `M`
            (=columns of `A`) equals the number of parameters
            (=elements of `x`)
        y : numpy.ndarray
            vector of target values

        Returns
        -------
        float
            root mean squared error

        """
        y_predicted = self.predict(A)
        delta_y = y_predicted - y
        rmse = np.sqrt(np.mean(delta_y**2))
        return rmse

    def predict(self, A):
        """
        Predicts data given an input matrix :math:`A`, i.e., :math:`Ax`,
        where :math:`x` is the vector of the fitted parameters.

        Parameters
        ----------
        A : numpy.ndarray
            fit matrix where `N` (=rows of `A`, elements of `y`) equals the
            number of target values and `M` (=columns of `A`) equals the number
            of parameters

        Returns
        -------
        numpy.ndarray or float
            vector of predicted values; float if single row provided as input

        """
        return np.dot(A, self.parameters)

    def get_contributions(self, A):
        """
        Computes the average contribution to the predicted values from each
        element of the parameter vector.

        Parameters
        ----------
        A : numpy.ndarray
            fit matrix where `N` (=rows of `A`, elements of `y`) equals the
            number of target values and `M` (=columns of `A`) equals the number
            of parameters

        Returns
        -------
        numpy.ndarray
            average contribution for each row of `A` from each parameter
        """
        return np.mean(np.abs(np.multiply(A, self.parameters)), axis=0)

    @abstractmethod
    def train(self):
        pass

    @property
    def summary(self):
        """ dict : comprehensive information about the optimizer """
        info = dict()
        info['seed'] = self.seed
        info['fit_method'] = self.fit_method
        info['standardize'] = self.standardize
        info['n_target_values'] = self.n_target_values
        info['n_parameters'] = self.n_parameters
        info['n_nonzero_parameters'] = \
            self.n_nonzero_parameters
        return {**info, **self._fit_results}

    def __str__(self):
        width = 54
        s = []
        s.append(' {} '.format(self.__class__.__name__).center(width, '='))
        for key in sorted(self.summary.keys()):
            value = self.summary[key]
            if isinstance(value, (str, int)):
                s.append('{:30} : {}'.format(key, value))
            elif isinstance(value, (float)):
                s.append('{:30} : {:.7g}'.format(key, value))
        s.append(''.center(width, '='))
        return '\n'.join(s)

    def __repr__(self):
        return 'BaseOptimizer((A, y), {}, {}'.format(
            self.fit_method, self.seed)

    @property
    def fit_method(self):
        """ str : fit method """
        return self._fit_method

    @property
    def parameters(self):
        """ numpy.ndarray : copy of parameter vector """
        if self._fit_results['parameters'] is None:
            return None
        else:
            return self._fit_results['parameters'].copy()

    @property
    def n_nonzero_parameters(self):
        """ int : number of non-zero parameters """
        if self.parameters is None:
            return None
        else:
            return np.count_nonzero(self.parameters)

    @property
    def n_target_values(self):
        """ int : number of target values (=rows in `A` matrix) """
        return self._n_rows

    @property
    def n_parameters(self):
        """ int : number of parameters (=columns in `A` matrix) """
        return self._n_cols

    @property
    def standardize(self):
        """ bool : whether or not to standardize the fit matrix before
        fitting
        """
        return self._standarize

    @property
    def seed(self):
        """ int : seed used to initialize pseudo random number generator """
        return self._seed



"""
This module contains the OrientationFamily class, which acts as a
container for information concernign a set of clusters of equal
orientation in space.
"""
import pickle


class OrientationFamily:
    """A container for storing information for a "family of orientations".

    An orbit contains many clusters. Some of the clusters can be tranlsated
    onto each other and other must first be rotated. A set of clusters in the
    orbit which can all be translated onto each other are oriented in the same
    way and belongs to the same orientation family. The family is haracterized
    by the symmetry (rotation) which relates it to the prototype structure of
    the orbit.

    Since the clusters are generally stored sorted the permutation information
    must also be stored.

    Parameters
    ----------
    symmetry_index : int
        The index of the symmetry corresponding to spglibs symmetry

    Attributes
    ----------
    symmetry_index : int
        The index of the symmetry corresponding to spglibs symmetry
    cluster_indices : list of ints
        The indices of the clusters belonging to this family
    permutation_indices : list of ints
        The indices of the permutation vector
    """

    def __init__(self, symmetry_index=None):
        self.symmetry_index = symmetry_index
        self.cluster_indices = []
        self.permutation_indices = []

    def write(self, f):
        """ Write the object to file.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to write to (file object)
        """
        try:
            pickle.dump(self, f)
        except Exception:
            raise Exception('Failed writing to file.')

    @staticmethod
    def read(f):
        """Load a OrientationFamily object from a pickle file.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to load from (file object)

        Returns
        -------
        OrientationFamily object
        """
        try:
            return pickle.load(f)
        except Exception:
            raise Exception('Failed loading from file.')



import pickle
import numpy as np
from ase.neighborlist import NeighborList


# TODO: Use another base class to hide internals from user
class Cutoffs:
    """ This class maintains information about the cutoff configuration,
    i.e. which clusters will be included (="inside cutoff"). It also
    encapsulates functionality that is used e.g., during cluster space
    construction.

    Here, `n-body` refers to number of atoms in a cluster. For example
    the cluster (0011) is a two-body cluster of fourth order and the
    cluster (123) is a three-body cluster of third order.

    Parameters
    ----------
    cutoff_matrix : numpy.ndarray
        the matrix element `ij` provides to the cutoff for order `i+2`
        and nbody `j+2`; elements with `j>i` will be ignored
    """

    def __init__(self, cutoff_matrix):

        self._cutoff_matrix = np.array(cutoff_matrix)

        for i in range(len(self._cutoff_matrix) - 1):
            assert max(self._cutoff_matrix[i, :]) >= \
                   max(self._cutoff_matrix[i+1, :]),\
                   'Make sure cutoffs are in decreasing order'

    @property
    def cutoff_matrix(self):
        """ numpy.ndarray : copy of cutoff matrix """
        return self._cutoff_matrix.copy()

    @property
    def orders(self):
        """ list(int) : allowed orders """
        return list(range(2, self.max_order + 1))

    @property
    def nbodies(self):
        """ list(int) : allowed bodies """
        return list(range(2, self.max_nbody + 1))

    def get_cutoff(self, order, nbody):
        """
        Returns cutoff for a given body and order.

        Parameters
        ----------
        order : int
        nbody : int

        Raises
        ------
        ValueError
            if order is not in orders
        ValueError
            if nbody is not in nbodies
        ValueError
            if nbody is larger than order

        Returns
        -------
        float
        """
        if order not in self.orders:
            raise ValueError('order not in orders')
        if nbody not in self.nbodies:
            raise ValueError('nbody not in nbodies')
        if nbody > order:
            raise ValueError('nbody can not be larger than order')
        return self._cutoff_matrix[nbody - 2, order - 2]

    @property
    def max_cutoff(self):
        """ float : maximum cutoff """
        max_cutoff = 0
        for i, row in enumerate(self._cutoff_matrix):
            max_cutoff = max(max_cutoff, np.max(row[i:]))
        return max_cutoff

    @property
    def max_nbody(self):
        """ int : maximum body """
        return self._cutoff_matrix.shape[0] + 1

    @property
    def max_order(self):
        """ int : maximum order """
        return self._cutoff_matrix.shape[1] + 1

    def max_nbody_cutoff(self, nbody):
        """ Return maximum cutoff for a given body. """
        if nbody not in self.nbodies:
            raise ValueError('nbody not in nbodies')
        return np.max(self._cutoff_matrix[nbody - 2, max(0, nbody - 2):])

    def max_nbody_order(self, nbody):
        """ Returns maximum order for a given body """
        if nbody not in self.nbodies:
            raise ValueError('nbody not in nbodies')
        arr = self._cutoff_matrix[nbody - 2, max(0, nbody - 2):]
        return np.count_nonzero(arr) + (nbody - 1)

    def write(self, fileobj):
        """ Writes instance to file.

        Parameters
        ----------
        fileobj : file-like object
            file-like object to which the cutoffs will be written to
        """
        pickle.dump(self._cutoff_matrix, fileobj)

    def read(fileobj):
        """ Reads an instance from file.

        Parameters
        ----------
        fileobj : file-like object
            input file to read from
        """
        data = pickle.load(fileobj)
        if type(data) is np.ndarray:
            return Cutoffs(data)
        else:
            cutoffs = data['cutoffs']
            return CutoffMaximumBody(cutoffs, len(cutoffs) + 1)

    def __str__(self):
        return str(self._cutoff_matrix)


class CutoffMaximumBody(Cutoffs):
    """ Specify cutoff-list plus maximum body

    Usefull when creating e.g. 6th order expansions but with only 3-body
    interactions.

    Parameters
    ----------
    cutoff_list : list
        list of cutoffs for order 2, 3, etc. Must be in decresing order
    max_nbody : int
        No clusters containing more than max_nbody atoms will be generated
    """

    def __init__(self, cutoff_list, max_nbody):
        cutoff_matrix = np.zeros((max_nbody - 1, len(cutoff_list)))
        for order, cutoff in enumerate(cutoff_list, start=2):
            cutoff_matrix[:, order - 2] = cutoff
        super().__init__(cutoff_matrix)


def is_cutoff_allowed(atoms, cutoff):
    """ Checks if atoms is compatible with cutoff

    Parameters
    ----------
    atoms : ase.Atoms
        structure used for checking compatibility with cutoff
    cutoff : float
        cutoff to be tested

    Returns
    -------
    bool
        True if cutoff compatible with atoms object, else False
    """
    nbrlist = NeighborList(cutoffs=[cutoff / 2] * len(atoms), skin=0,
                           self_interaction=False, bothways=True)
    nbrlist.update(atoms)

    for i in range(len(atoms)):
        neighbors, _ = nbrlist.get_neighbors(i)
        if i in neighbors:
            return False
        if len(neighbors) != len(set(neighbors)):
            return False
    return True


def estimate_maximum_cutoff(atoms, max_iter=11):
    """ Estimates the maximum possible cutoff given the atoms object

    Parameters
    ----------
    atoms : ase.Atoms
        structure used for checking compatibility with cutoff
    max_iter : int
        number of iterations in binary search
    """

    # First upper boundary of cutoff
    upper_cutoff = min(np.linalg.norm(atoms.cell, axis=1))

    # generate all possible offsets given upper_cutoff
    nbrlist = NeighborList(cutoffs=[upper_cutoff / 2] * len(atoms), skin=0,
                           self_interaction=False, bothways=True)
    nbrlist.update(atoms)
    all_offsets = []
    for i in range(len(atoms)):
        _, offsets = nbrlist.get_neighbors(i)
        all_offsets.extend([tuple(offset) for offset in offsets])

    # find lower boundary and new upper boundary
    unique_offsets = set(all_offsets)
    unique_offsets.discard((0, 0, 0))
    upper = min(np.linalg.norm(np.dot(offset, atoms.cell))
                for offset in unique_offsets)
    lower = upper / 2.0

    # run binary search between the upper and lower bounds
    for _ in range(max_iter):
        cutoff = (upper + lower) / 2
        if is_cutoff_allowed(atoms, cutoff):
            lower = cutoff
        else:
            upper = cutoff
    return lower



from itertools import product
import numpy as np


def _obj2str(a, none_char='-'):
    """ Casts object a to str. """
    if isinstance(a, float):
        # if the float is 2.49999999 then round
        if str(a)[::-1].find('.') > 5:
            return '{:.5f}'.format(a)
    elif a is None:
        return none_char
    return str(a)


_array2str = np.vectorize(_obj2str)


def print_table(matrix, sum_=False):
    """ Prints matrix data in a nice table format.

    The matrix element matrix[i][j] should correspond to information about
    order j+2 and n-body i+1.

    Example
    --------
    >> matrix = numpy.array([[None, None], [4.0, 3.0]])
    >> print_table(matrix)

     body/order |  2  |  3
    ------------------------
         1      |  -  |  -
         2      | 4.0 | 3.0

    Parameters
    ----------
    matrix : numpy.ndarray
        matrix to be printed
    sum: : bool
        whether or not to print the sum along each row and column
    """

    # initialze table
    n_rows, n_cols = matrix.shape
    A = _build_table_frame(order=n_cols+1, nbody=n_rows, sum_=sum_)

    # fill table
    for order, nbody in product(range(2, n_cols+2), range(1, n_rows+1)):
        if nbody <= order:
            A[nbody, order-1] = matrix[nbody-1, order-2]

    if sum_:
        for i, row in enumerate(A[1:-1, 1:-1], start=1):
            A[i, -1] = sum(val for val in row if val is not None)
        for i, col in enumerate(A[1:-1, 1:-1].T, start=1):
            A[-1, i] = sum(val for val in col if val is not None)
        A[-1, -1] = ''

    _print_table(_array2str(A))


def _build_table_frame(order, nbody, sum_=False):
    """ Builds/initializes table/array. """
    if sum_:
        A = np.empty((nbody+2, order+1), dtype='object')
        A[0, -1] = 'sum'
        A[-1, 0] = 'sum'
    else:
        A = np.empty((nbody+1, order), dtype='object')

    A[0][0] = 'body/order'
    A[0, 1:order] = range(2, order+1)
    A[1:nbody+1, 0] = range(1, nbody+1)
    return A


def _print_table(table):
    """ Prints 2D array with str elements. """
    n_rows, n_cols = table.shape

    # find maximum widths for each column
    widths = []
    for i in range(n_cols):
        widths.append(max(len(val) for val in table[:, i])+2)

    # formatting str for each row
    row_format = '|'.join('{:^'+str(width)+'}' for width in widths)

    # print
    for i in range(n_rows):
        if i == 1:
            print('-'*(sum(widths)+n_rows-1))
        print(row_format.format(*table[i, :]))


if __name__ == "__main__":
    # input dummy cutoff table
    # insert row for nbody=1
    cutoffs = np.array([[None, None, None, None, None],
                        [6.0, 6.0, 6.0, 3.7, 3.7],
                        [5.0, 5.0, 5.0, 3.0, 3.0],
                        [3.7, 3.7, 3.7, 0.0, 0.0]])

    # input dummy cluster count table
    cluster_counts = np.array([[1, 3, 5, 5, 2],
                               [12, 22, 39, 42, 58],
                               [19, 41, 123, 421, 912],
                               [42, 112, 410, 617, 3271]])

    print_table(cutoffs)
    print('\n')
    print_table(cluster_counts, sum_=False)
    print('\n')
    print_table(cluster_counts, sum_=True)



"""
This module implements the split-Bregman algorithm described in
T. Goldstein and S. Osher, SIAM J. Imaging Sci. 2, 323 (2009);
doi:10.1137/080725891
"""

import numpy as np
from scipy.optimize import minimize


def fit_split_bregman(A, y, mu=1e-3, lmbda=100, n_iters=1000, tol=1e-6,
                      verbose=0):
    """
    Split-Bregman algorithm described in T. Goldstein and S. Osher,
    SIAM J. Imaging Sci. 2, 323 (2009); doi:10.1137/080725891.

    Parameters
    -----------
    X : numpy.ndarray
        fit matrix
    y : numpy.ndarray
        target array
    mu : float
        Sparseness parameter
    lmbda : float
        Split Bregman parameter
    n_iters : int
        maximal number of split bregman iterations.
    tol : float
        tolerance for when stopping split bregman iterations.

    Returns
    ----------
    results : dict
        parameters
    """

    n_cols = A.shape[1]
    d = np.zeros(n_cols)
    b = np.zeros(n_cols)
    x = np.zeros(n_cols)

    old_norm = 0.0

    # Precompute for speed.
    AtA = np.dot(A.conj().transpose(), A)
    ftA = np.dot(y.conj().transpose(), A)
    ii = 0
    for i in range(n_iters):
        if verbose:
            print('Iteration ', i)
        args = (A, y, mu, lmbda, d, b, AtA, ftA)
        res = minimize(_objective_function, x, args, method='BFGS', options={
                       'disp': False}, jac=_objective_function_derivative)
        x = res.x

        d = _shrink(mu*x + b, 1.0/lmbda)
        b = b + mu*x - d

        new_norm = np.linalg.norm(x)
        ii = ii + 1

        if verbose:
            print('|new_norm-old_norm| = ', abs(new_norm-old_norm))
        if abs(new_norm-old_norm) < tol:
            break

        old_norm = new_norm
    else:
        print('Warning: Split Bregman ran for max iters')

    fit_results = {'parameters': x}
    return fit_results


def _objective_function(x, A, y, mu, lmbda, d, b, AtA, ftA):
    """ Objective function to be minimized.

    Parameters
    -----------
    X : numpy.ndarray
        fit matrix
    y : numpy.ndarray
        target array
    mu : float
        the parameter that adjusts sparseness.
    lmbda : float
        Split Bregman parameter
    d : numpy.ndarray
        same notation as Nelson, Hart paper
    b : numpy.ndarray
        same notation as Nelson, Hart paper
    AtA : numpy.ndarray
        sensing matrix transpose times sensing matrix.
    ftA : numpy.ndarray
        np.dot(y.conj().transpose(), A)
    """

    error_vector = np.dot(A, x) - y

    obj_function = 0.5*np.vdot(error_vector, error_vector)

    if obj_function.imag > 0.0:
        raise RuntimeError(
            'Objective function contains non-zero imaginary part.)')

    sparseness_correction = d - b - mu*x
    obj_function += 0.5*lmbda * \
        np.vdot(sparseness_correction, sparseness_correction)

    if obj_function.imag > 0.0:
        raise RuntimeError(
            'Objective function contains non-zero imaginary part.)')

    return obj_function


def _objective_function_derivative(x, A, y, mu, lmbda, d, b, AtA, ftA):
    """ Derivative of the objective function.

    Parameters
    -----------
    X : numpy.ndarray
        fit matrix
    y : numpy.ndarray
        target array
    mu : float
        the parameter that adjusts sparseness.
    lmbda : float
        Split Bregman parameter
    d : numpy.ndarray
        same notation as Nelson, Hart paper
    b : numpy.ndarray
        same notation as Nelson, Hart paper
    AtA : numpy.ndarray
        sensing matrix transpose times sensing matrix.
    ftA : numpy.ndarray
        np.dot(y.conj().transpose(), A)

    """
    ret = np.squeeze(np.dot(x[np.newaxis, :], AtA) -
                     ftA - lmbda*mu*(d - mu * x - b))
    return ret


def _shrink(y, alpha):
    """
    Shrink operator as defined in Eq. (11) (p. 5)
    in Nelson, Hart (Compressive sensing as a new
    paradigm for model building).

    Parameters
    -----------
    y : numpy.ndarray
    alpha : float
    """
    return np.sign(y) * np.maximum(np.abs(y) - alpha, 0.0)



"""
Contains the Orbit class which hold onformation about equivalent clusters.
"""

import numpy as np
import tarfile

from .orientation_family import OrientationFamily
from ..io.read_write_files import (add_items_to_tarfile_hdf5,
                                   add_items_to_tarfile_pickle,
                                   add_list_to_tarfile_custom,
                                   read_items_hdf5,
                                   read_items_pickle,
                                   read_list_custom)


class Orbit:
    """
    This class serves as a container for storing data pertaining to an orbit.

    Attributes
    ----------
    orientation_families : list of OrientationFamily objs
        orientation families of the orbit
    eigensymmetries : list of tuples
        each eigensymmetry corresponds to a pair where the first index
        is the symmetry and the second is the permutation
    eigentensors : list(numpy.ndarray)
        decomposition of the force constant into symmetry elements
    """

    # TODO: Make properties of the parameters
    def __init__(self):
        self.orientation_families = []
        self.eigensymmetries = []
        self.eigentensors = []

    @property
    def prototype_index(self):
        """int : index of cluster that serves as prototype for this orbit

        In the code the first symmetry is always the identity so the first
        orientation family should always correspond to the prototype"""
        return self.orientation_families[0].cluster_indices[0]

    def write(self, f):
        """Write a Orbit instance to a file.

        Parameters
        ----------
        f : str or file object
            name of input file (str) or stream to write to (file object)
        """
        tar_file = tarfile.open(mode='w', fileobj=f)

        # add eigentensors as NumPy array
        items_hdf5 = dict(eigentensors=np.array(self.eigentensors),
                          )
        add_items_to_tarfile_hdf5(tar_file, items_hdf5, 'eigentensors')

        # add eigensymmetries as list
        items_pickle = dict(eigensymmetries=self.eigensymmetries,
                            order=self.order, radius=self.radius,
                            maximum_distance=self.maximum_distance)
        add_items_to_tarfile_pickle(tar_file, items_pickle, 'attributes')

        # add orientation families
        add_list_to_tarfile_custom(tar_file, self.orientation_families,
                                   'orientation_families')
        tar_file.close()

    @staticmethod
    def read(f):
        """Load a ClusterSpace from file

        Parameters
        ----------
        f : string or file object
            name of input file (string) or stream to load from (file object)
        """

        orb = Orbit()
        tar_file = tarfile.open(mode='r', fileobj=f)

        # read eigentensors hdf5
        items_hdf5 = read_items_hdf5(tar_file, 'eigentensors')
        orb.eigentensors = [et for et in items_hdf5['eigentensors']]

        # read attributes pickle
        attributes = read_items_pickle(tar_file, 'attributes')
        for name, value in attributes.items():
            orb.__setattr__(name, value)

        # read orientation families
        ofs = read_list_custom(tar_file, 'orientation_families',
                               OrientationFamily.read)
        orb.orientation_families = ofs
        return orb


def get_geometrical_radius(positions):
    """Compute the geometrical size of a 3-dimensional point cloud. The
    geometrical size is defined as the average distance to the geometric
    center.

    Parameters
    ----------
    positions : list of 3-dimensional vectors
        positions of points in cloud

    Returns
    -------
    float
        geometric size of point cloud
    """
    geometric_center = np.mean(positions, axis=0)
    return np.mean(np.sqrt(np.sum((positions - geometric_center)**2, axis=1)))


def get_maximum_distance(positions):
    """Compute the maximum distance between any two points in a 3-dimensional
    point cloud. This is equivalent to the "size" criterion used when imposing
    a certain (pair) cutoff criterion during construction of a set of clusters.

    Parameters
    ----------
    positions : list of 3-dimensional vectors
        positions of points in cloud

    Returns
    -------
    float
        maximum distance betwee any two points
    """
    if len(positions) == 0:
        return 0.0
    max_distance = 0.0
    for pt1 in positions[:-1]:
        for pt2 in positions[1:]:
            max_distance = max(max_distance, np.linalg.norm(pt1 - pt2))
    return max_distance
